﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using Microsoft.Xna.Framework;

// db
using Amazon;
using Amazon.SimpleDB;
using Amazon.SimpleDB.Util;
using Amazon.SimpleDB.Model;

namespace DBParameterEditor
{
    public class ParameterProcessor
    {
        #region singleton
        // create singleton
        static readonly ParameterProcessor instance = new ParameterProcessor();
        /// <summary>
        /// Get instance of singleton.
        /// </summary>
        public static ParameterProcessor Instance
        {
            get
            {
                return instance;
            }
        }
        /// <summary>
        /// Explicit static constructor to tell C# compiler
        /// not to mark type as beforefieldinit
        /// </summary>
        static ParameterProcessor()
        {
        }
        #endregion

        // db
        AmazonSimpleDB m_sdb;

        /// <summary>
        /// Constructor
        /// </summary>
        ParameterProcessor()
        {
            // intitialize db
            // amazon cloud access
            m_sdb = AWSClientFactory.CreateAmazonSimpleDBClient("YourPublicKey", "YourPrivatKey");
        }
        public void SaveStructToFile<T>(T data, string filename)
        {
            // Delete the file if it exists.
            if (File.Exists(filename))
            {
                File.Delete(filename);
            }

            using (FileStream fileStream = File.Create(filename))
            {
                XmlWriterSettings writerSettings =
                new XmlWriterSettings
                {
                    Indent = true,
                    IndentChars = "\t"
                };

                XmlSerializer serializer = new XmlSerializer(typeof(T));


                using (XmlWriter xmlWriter = XmlWriter.Create(fileStream, writerSettings))
                {
                    serializer.Serialize(xmlWriter, data);
                }
            }

        }
        /// <summary>
        /// Takes a class and enters its fields to an Amazon SimpleDB
        /// </summary>
        /// <typeparam name="T">Serializable Class to be saved into the database</typeparam>
        /// <param name="data">Class instance that is saved into the database</param>
        /// <param name="tableName">Name of domain in SimpleDB</param>
        public void SaveToDB<T>(T data, string tableName)
        {
            SimpleDBResponseEventHandler<object, ResponseEventArgs> handler = null;
            handler = delegate(object senderAmazon, ResponseEventArgs args)
            {
                //Unhook from event.
                m_sdb.OnSimpleDBResponse -= handler;
                PutAttributesResponse response = args.Response as PutAttributesResponse;
                if (null != response)
                {
                }
                else
                {
                }
            };

            m_sdb.OnSimpleDBResponse += handler;

            string structName = data.GetType().Name;
            PutAttributesRequest putAttributesRequest = new PutAttributesRequest { DomainName = tableName, ItemName = structName };
            List<ReplaceableAttribute> attributesOne = putAttributesRequest.Attribute;

            Type type = data.GetType();
            System.Reflection.FieldInfo[] fields = type.GetFields();
            foreach (System.Reflection.FieldInfo field in fields)
            {
                string valueString = FieldToString(field, data);
                attributesOne.Add(new ReplaceableAttribute().WithName(field.Name).WithValue(valueString).WithReplace(true));
            }
            attributesOne.Add(new ReplaceableAttribute().WithName("TimeStamp").WithValue(DateTime.Now.Ticks.ToString()).WithReplace(true));

            m_sdb.PutAttributes(putAttributesRequest);
        }

        /// <summary>
        /// Reads values from a SimpleDB and writes it into a class instance.
        /// </summary>
        /// <typeparam name="T">Serializable Class that is read from the database</typeparam>
        /// <param name="data">Class instance where the database values are written into</param>
        /// <param name="tableName">Name of domain in SimpleDB</param>
        /// <param name="readFinished">A delegate that is called when reading the values has finished. Set null if not needed.</param>
        public void ReadFromDB<T>(T data, string tableName, Action readFinished)
        {
            // create response handler and delegate
            SimpleDBResponseEventHandler<object, ResponseEventArgs> handler = null;

            handler = delegate(object sender, ResponseEventArgs args)
            {
                //Unhook from event.
                m_sdb.OnSimpleDBResponse -= handler;
                SelectResponse response = args.Response as SelectResponse;

                if (null != response)
                {
                    Type type = data.GetType();
                    System.Reflection.FieldInfo[] fields = type.GetFields();

                    SelectResult selectResult = response.SelectResult;
                    if (null != selectResult)
                    {
                        foreach (Item item in selectResult.Item)
                        {
                            // actually should just be one item
                            foreach (Amazon.SimpleDB.Model.Attribute attribute in item.Attribute)
                            {
                                var field = fields.FirstOrDefault(f => f.Name == attribute.Name);
                                if (field != null)
                                {
                                    StringToFieldValue(field, data, attribute.Value);
                                }
                            }
                        }
                    }
                    if (readFinished != null)
                    {
                        readFinished.Invoke();
                    }
                }
            };
            m_sdb.OnSimpleDBResponse += handler;

            string structName = data.GetType().Name;
            // create request
            string sql = "SELECT * FROM " + tableName + " WHERE itemName()='" + structName + "'";
            m_sdb.Select(new SelectRequest { SelectExpression = sql, ConsistentRead = true });
        }

        private void DeleteItem(string itemName, string tableName, List<Amazon.SimpleDB.Model.Attribute> attributes)
        {
            SimpleDBResponseEventHandler<object, ResponseEventArgs> handler = null;
            handler = delegate(object senderAmazon, ResponseEventArgs args)
            {
                //Unhook from event.
                m_sdb.OnSimpleDBResponse -= handler;
                PutAttributesResponse response = args.Response as PutAttributesResponse;
                if (null != response)
                {
                }
                else
                {
                }
            };

            m_sdb.OnSimpleDBResponse += handler;

            DeleteAttributesRequest deleteAttributesRequest = new DeleteAttributesRequest { DomainName = tableName, ItemName = itemName };
            List<Amazon.SimpleDB.Model.Attribute> attributesOne = deleteAttributesRequest.Attribute;

            foreach (Amazon.SimpleDB.Model.Attribute attr in attributes)
            {
                attributesOne.Add(attr);
            }

            m_sdb.DeleteAttributes(deleteAttributesRequest);
        }

        /// <summary>
        /// Delete the entry of a complete item (data class). Call this whenever you changed the class fields, to avoid
        /// having fields in the database without corresponding fields in the class.
        /// </summary>
        /// <typeparam name="T">Serializable Class that is deleted from the database</typeparam>
        /// <param name="data">Class instance to be deleted from the database</param>
        /// <param name="tableName">Name of domain in SimpleDB</param>
        public void DeleteItem<T>(T data, string tableName)
        {
            // simpleDB does not have a delete item function.
            // to completely delete an item all attributes have to be deletet
            // therefore this is an asynchronous two step task
            // first request all attributes
            // then delete all attributes
            string itemName = data.GetType().Name;

            // create response handler and delegate
            SimpleDBResponseEventHandler<object, ResponseEventArgs> handler = null;

            handler = delegate(object sender, ResponseEventArgs args)
            {
                //Unhook from event.
                m_sdb.OnSimpleDBResponse -= handler;
                SelectResponse response = args.Response as SelectResponse;

                if (null != response)
                {
                    SelectResult selectResult = response.SelectResult;
                    if (null != selectResult)
                    {
                        DeleteItem(itemName, tableName, selectResult.Item[0].Attribute);
                    }
                }
            };
            m_sdb.OnSimpleDBResponse += handler;

            // create request
            string sql = "SELECT * FROM " + tableName + " WHERE itemName()='" + itemName + "'";
            m_sdb.Select(new SelectRequest { SelectExpression = sql, ConsistentRead = true });
        }

        /// <summary>
        /// Using reflection to get a value string of a class field.
        /// </summary>
        /// <param name="field">The field of the class to get the value from</param>
        /// <param name="data">Class instance with values</param>
        /// <returns></returns>
        public String FieldToString(System.Reflection.FieldInfo field, object data)
        {
            string valueString;
            if (field.FieldType == typeof(float))
            {
                valueString = ((float)field.GetValue(data)).ToString(System.Globalization.CultureInfo.InvariantCulture.NumberFormat);
            }
            else if (field.FieldType == typeof(Vector2))
            {
                Vector2 vec = (Vector2)field.GetValue(data);

                System.Globalization.NumberFormatInfo nfi
                        = System.Globalization.CultureInfo.InvariantCulture.NumberFormat;
                valueString = vec.X.ToString(nfi) + ", "
                            + vec.Y.ToString(nfi);
            }
            else
            {
                // just use the default ToString for all other field types
                // extend it if you need special handling like the types before
                valueString = field.GetValue(data).ToString();
            }
            return valueString;
        }

        /// <summary>
        /// Using reflection to set a class field from a value string.
        /// Currently only float, Vector2 and int is supported.
        /// Extend for you needs.
        /// </summary>
        /// <param name="field">The field of the class for which the value will be set.</param>
        /// <param name="data">Class instance where the value will be stored into</param>
        /// <param name="valueString">A string containing a value for the field</param>
        public void StringToFieldValue(System.Reflection.FieldInfo field, object data, string valueString)
        {
            if (field.FieldType == typeof(float))
            {
                field.SetValue(data, float.Parse(valueString, System.Globalization.CultureInfo.InvariantCulture.NumberFormat));
            }
            else if (field.FieldType == typeof(Vector2))
            {
                System.Globalization.NumberFormatInfo nfi
                    = System.Globalization.CultureInfo.InvariantCulture.NumberFormat;
                String[] values = valueString.Split(',');
                Vector2 point = new Vector2();
                try
                {
                    point.X = float.Parse(values[0], nfi);
                    point.Y = float.Parse(values[1], nfi);
                }
                catch
                {
                    // oh no
                }
                field.SetValue(data, point);
            }
            if (field.FieldType == typeof(int))
            {
                field.SetValue(data, Int32.Parse(valueString, System.Globalization.CultureInfo.InvariantCulture.NumberFormat));
            }
            else
            {
                // unsupported type
            }
        }
    }
}
